@@ -5,9 +5,12 @@ import configparser
 from collections import defaultdict
 from .remotejit import RemoteJIT
 from .thrift.utils import resolve_includes
-from pymapd.cursor import make_row_results_set
-from pymapd._parsers import _extract_description  # , _bind_parameters
-from .omnisci_backend import array_type_converter, OmnisciCompilerPipeline
+from .omnisci_backend import (
+    array_type_converter,
+    output_column_type_converter, column_type_converter,
+    table_function_sizer_type_converter,
+    OmnisciOutputColumnType, OmnisciColumnType,
+    OmnisciCompilerPipeline)
 from .targetinfo import TargetInfo
 from .irtools import compile_to_LLVM
 from .errors import ForbiddenNameError
@@ -33,7 +36,7 @@ def is_available(_cache={}):
                 _cache['reason'] = (
                     'expected OmniSci version 4.6 or greater, got %s'
                     % (version,))
-    return _cache.get('version', ()), _cache.get('reason')
+    return _cache.get('version', ()), _cache.get('reason', '')
 
 
 def get_client_config(**config):
@@ -149,7 +152,9 @@ class RemoteOmnisci(RemoteJIT):
     Use pymapd, for instance, to make a SQL query `select add(c1,
     c2) from table`
     """
-    converters = [array_type_converter]
+    converters = [array_type_converter,
+                  output_column_type_converter, column_type_converter,
+                  table_function_sizer_type_converter]
     multiplexed = False
     mangle_prefix = ''
 
@@ -254,7 +259,135 @@ class RemoteOmnisci(RemoteJIT):
             print(msg)
         return client(Omnisci={name: args})['Omnisci'][name]
 
+    def get_tables(self):
+        """Return a list of table names stored in the OmnisciDB server.
+        """
+        return self.thrift_call('get_tables', self.session_id)
+
+    def get_table_details(self, table_name):
+        """Return details about OmnisciDB table.
+        """
+        return self.thrift_call('get_table_details',
+                                self.session_id, table_name)
+
+    def load_table_columnar(self, table_name, **columnar_data):
+        """Load columnar data to OmnisciDB table.
+
+        Warning: when connected to OmnisciDB < 5.3, the data is loaded
+          row-wise that will be very slow for large data sets.
+
+        Parameters
+        ----------
+        table_name : str
+          The name of table into the data is loaded. The table must exist.
+
+        columnar_data : dict
+          A dictionary of column names and the corresponding column
+          values. A column values is a sequence of values.
+          The table must have the specified columns defined.
+
+        """
+        int_col_types = ['TINYINT', 'SMALLINT', 'INT', 'BIGINT', 'BOOL',
+                         'DECIMAL', 'TIME', 'TIMESTAMP', 'DATE']
+        real_col_types = ['FLOAT', 'DOUBLE']
+        str_col_types = ['STR', 'POINT', 'LINESTRING', 'POLYGON',
+                         'MULTIPOLYGON']
+
+        thrift = self.thrift_client.thrift
+        table_details = self.get_table_details(table_name)
+        if self.version[:2] < (5, 3):
+            rows = None
+            for column_name, column_data in columnar_data.items():
+                if rows is None:
+                    rows = [[None for j in range(len(columnar_data))]
+                            for i in range(len(column_data))]
+                col_index = None
+                for i, ct in enumerate(table_details.row_desc):
+                    if ct.col_name == column_name:
+                        typeinfo = ct.col_type
+                        col_index = i
+                        datumtype = thrift.TDatumType._VALUES_TO_NAMES[
+                            typeinfo.type]
+                        is_array = typeinfo.is_array
+                        break
+                assert col_index is not None
+                for i, v in enumerate(column_data):
+                    if is_array:
+                        if datumtype == 'BOOL':
+                            v = ["'true'" if v_ else "'false'" for v_ in v]
+                        v = ', '.join(map(str, v))
+                        v = f'ARRAY[{v}]'
+                    else:
+                        if datumtype == 'BOOL':
+                            v = "'true'" if v else "'false'"
+                    rows[i][col_index] = v
+
+            for row in rows:
+                table_row = ', '.join(map(str, row))
+                self.sql_execute(
+                    f'INSERT INTO {table_name} VALUES ({table_row})')
+            return
+
+        columns = []
+        for column_name, column_data in columnar_data.items():
+            typeinfo = None
+            for ct in table_details.row_desc:
+                if ct.col_name == column_name:
+                    typeinfo = ct.col_type
+                    break
+            if typeinfo is None:
+                raise ValueError(
+                    f'OmnisciDB `{table_name}` has no column `{column_name}`')
+            datumtype = thrift.TDatumType._VALUES_TO_NAMES[typeinfo.type]
+            if typeinfo.is_array:
+                rows = []
+                for row in column_data:
+                    if datumtype in int_col_types:
+                        row_data = thrift.TColumnData(int_col=row)
+                    elif datumtype in real_col_types:
+                        row_data = thrift.TColumnData(real_col=row)
+                    elif datumtype in str_col_types:
+                        row_data = thrift.TColumnData(str_col=row)
+                    else:
+                        raise NotImplementedError(
+                            f'loading {datumtype} array data')
+                    rows.append(thrift.TColumn(
+                        data=row_data,
+                        nulls=[False] * len(row)))
+                col_data = thrift.TColumnData(arr_col=rows)
+            elif datumtype in int_col_types:
+                col_data = thrift.TColumnData(int_col=column_data)
+            elif datumtype in real_col_types:
+                col_data = thrift.TColumnData(real_col=column_data)
+            elif datumtype in str_col_types:
+                col_data = thrift.TColumnData(str_col=column_data)
+            else:
+                raise NotImplementedError(f'loading {datumtype} data')
+            columns.append(thrift.TColumn(
+                data=col_data, nulls=[False] * len(column_data)))
+
+        self.thrift_call('load_table_binary_columnar',
+                         self.session_id, table_name, columns)
+
     def sql_execute(self, query):
+        """Execute SQL query in OmnisciDB server.
+
+        Parameters
+        ----------
+        query : str
+
+          SQL query string containing exactly one query. Multiple
+          queries are not supported.
+
+        Returns
+        -------
+        descr : object
+          Row description object
+        results : iterator
+          Iterator over rows.
+        """
+        from pymapd.cursor import make_row_results_set
+        from pymapd._parsers import _extract_description
         self.register()
         columnar = True
         if self.debug:
@@ -294,6 +427,13 @@ class RemoteOmnisci(RemoteJIT):
             'Array<int64_t>': typemap['TExtArgumentType']['ArrayInt64'],
             'Array<float>': typemap['TExtArgumentType']['ArrayFloat'],
             'Array<double>': typemap['TExtArgumentType']['ArrayDouble'],
+            'Column<bool>': typemap['TExtArgumentType'].get('ColumnBool'),
+            'Column<int8_t>': typemap['TExtArgumentType'].get('ColumnInt8'),
+            'Column<int16_t>': typemap['TExtArgumentType'].get('ColumnInt16'),
+            'Column<int32_t>': typemap['TExtArgumentType'].get('ColumnInt32'),
+            'Column<int64_t>': typemap['TExtArgumentType'].get('ColumnInt64'),
+            'Column<float>': typemap['TExtArgumentType'].get('ColumnFloat'),
+            'Column<double>': typemap['TExtArgumentType'].get('ColumnDouble'),
             'Cursor': typemap['TExtArgumentType']['Cursor'],
             'void': typemap['TExtArgumentType']['Void'],
             'GeoPoint': typemap['TExtArgumentType'].get('GeoPoint'),
@@ -302,6 +442,11 @@ class RemoteOmnisci(RemoteJIT):
             'GeoMultiPolygon': typemap['TExtArgumentType'].get(
                 'GeoMultiPolygon'),
         }
+
+        if self.version[:2] < (5, 4):
+            ext_arguments_map['Array<bool>'] = typemap[
+                'TExtArgumentType']['ArrayInt8']
+
         ext_arguments_map['{bool* ptr, uint64 sz, bool is_null}*'] \
             = ext_arguments_map['Array<bool>']
         ext_arguments_map['{int8* ptr, uint64 sz, bool is_null}*'] \
@@ -316,6 +461,19 @@ class RemoteOmnisci(RemoteJIT):
             = ext_arguments_map['Array<float>']
         ext_arguments_map['{float64* ptr, uint64 sz, bool is_null}*'] \
             = ext_arguments_map['Array<double>']
+
+        for ptr_type, T in [
+                ('bool', 'bool'),
+                ('int8', 'int8_t'),
+                ('int16', 'int16_t'),
+                ('int32', 'int32_t'),
+                ('int64', 'int64_t'),
+                ('float32', 'float'),
+                ('float64', 'double'),
+        ]:
+            ext_arguments_map['{%s* ptr, uint64 sz}' % ptr_type] \
+                = ext_arguments_map.get('Column<%s>' % T)
+
         values = list(ext_arguments_map.values())
         for v, n in thrift.TExtArgumentType._VALUES_TO_NAMES.items():
             if v not in values:
@@ -393,9 +551,21 @@ class RemoteOmnisci(RemoteJIT):
             return
         thrift = self.thrift_client.thrift
 
+        sizer_map = dict(
+            ConstantParameter='kUserSpecifiedConstantParameter',
+            RowMultiplier='kUserSpecifiedRowMultiplier',
+            Constant='kConstant')
+
+        def is_udtf(sig):
+            for a in sig[1]:
+                if isinstance(a, (OmnisciOutputColumnType, OmnisciColumnType)):
+                    return True
+            return False
+
         udfs = []
         udtfs = []
         device_ir_map = {}
+        unspecified = object()
         for device, target_info in self.targets.items():
             functions_and_signatures = []
             function_signatures = defaultdict(list)
@@ -424,12 +594,74 @@ class RemoteOmnisci(RemoteJIT):
                                   f' in {f2}#{n2}.')
                         continue
                     function_signatures[name].append(sig)
-                    is_udtf = 'table' in sig[0].annotation()
-                    if i == 0 and (self.version < (5, 2) or is_udtf):
+
+                    sig_is_udtf = is_udtf(sig)
+                    is_old_udtf = 'table' in sig[0].annotation()
+                    if i == 0 and (self.version < (5, 2)
+                                   or is_old_udtf or sig_is_udtf):
+                        # TODO: support __0 for UDTFs
                         sig.set_mangling('')
                     else:
                         sig.set_mangling('__%s' % (i))
-                    if is_udtf:
+
+                    if sig_is_udtf:
+                        # new style UDTF, requires omniscidb version >= 5.4
+
+                        if self.version < (5, 4):
+                            v = '.'.join(map(str, self.version))
+                            raise RuntimeError(
+                                'UDTFs with Column arguments require '
+                                'omniscidb 5.4 or newer, currently '
+                                'connected to ', v)
+
+                        inputArgTypes = []
+                        outputArgTypes = []
+                        sqlArgTypes = []
+                        sizer = None
+                        sizer_index = -1
+                        for i, a in enumerate(sig[1]):
+                            annot = a.annotation()
+                            _sizer = annot.get('sizer', unspecified)
+                            if _sizer is not unspecified:
+                                if not (a.is_int and a.bits == 32):
+                                    raise ValueError(
+                                        'sizer argument must have type int32')
+                                if _sizer is None:
+                                    _sizer = 'RowMultiplier'
+                                _sizer = sizer_map[_sizer]
+                                # cannot have multiple sizer arguments
+                                assert sizer_index == -1
+                                sizer_index = i + 1
+                                sizer = _sizer
+                            atype = ext_arguments_map[
+                                a.tostring(use_annotation=False)]
+                            if isinstance(a, OmnisciOutputColumnType):
+                                outputArgTypes.append(atype)
+                            elif isinstance(a, OmnisciColumnType):
+                                sqlArgTypes.append(ext_arguments_map['Cursor'])
+                                inputArgTypes.append(atype)
+                            else:
+                                sqlArgTypes.append(atype)
+                                inputArgTypes.append(atype)
+
+                        assert sizer is not None  # need a sizer argument
+                        sizer_type = getattr(
+                            thrift.TOutputBufferSizeType, sizer)
+                        udtfs.append(thrift.TUserDefinedTableFunction(
+                            name + sig.mangling(),
+                            sizer_type, sizer_index,
+                            inputArgTypes, outputArgTypes, sqlArgTypes))
+
+                    elif is_old_udtf:
+                        # old style UDTF for omniscidb <= 5.3, to be deprecated
+
+                        if self.version >= (5, 4):
+                            v = '.'.join(map(str, self.version))
+                            raise RuntimeError(
+                                'Old-style UDTFs require '
+                                'omniscidb 5.3 or older, currently '
+                                'connected to ', v)
+
                         sizer = None
                         sizer_index = -1
                         inputArgTypes = []
@@ -458,7 +690,7 @@ class RemoteOmnisci(RemoteJIT):
                         sizer_type = getattr(
                             thrift.TOutputBufferSizeType, sizer)
                         udtfs.append(thrift.TUserDefinedTableFunction(
-                            name + sig.mangling,
+                            name + sig.mangling(),
                             sizer_type, sizer_index,
                             inputArgTypes, outputArgTypes, sqlArgTypes))
                     else:
@@ -467,7 +699,8 @@ class RemoteOmnisci(RemoteJIT):
                         atypes = [ext_arguments_map[a.tostring(
                             use_annotation=False)] for a in sig[1]]
                         udfs.append(thrift.TUserDefinedFunction(
-                            name + sig.mangling, atypes, rtype))
+                            name + sig.mangling(),
+                            atypes, rtype))
                     signatures.append(sig)
                 functions_and_signatures.append((caller.func, signatures))
             llvm_module = compile_to_LLVM(functions_and_signatures,
